from basil.utils import camel_to_snake_case


class Benchmark:

    @property
    def name(self):
        # Transform class name into snake case
        return camel_to_snake_case(self.__class__.__name__)

    def validate_outputs(self, outputs):
        pass



# This scripot has arguments
import argparse

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import streamlit as st

from basil.utils import compute_pareto_frontier, load_results

arg_parser = argparse.ArgumentParser(description='Process some inputs.')
arg_parser.add_argument('--folder',
                        type=str,
                        help='folder containing the benchmark runs')
args = arg_parser.parse_args()

st.title('Basil')
st.write(f'{args.folder}')

# Load all runs
result_families = load_results(args.folder)

# Converting the result_families to a pandas dataframe for easier manipulation
data = []
for family in result_families:
    for result in family.results:
        # Flatten the data
        flattened_result = {
            **result.system_info,
            **result.options,
            **result.parameters,
            **result.results
        }
        flattened_result["solver"] = result.solver
        flattened_result["name"] = family.name
        data.append(flattened_result)

df = pd.DataFrame(data)

# Filter for names
names = df['name'].unique().tolist()
selected_name = st.sidebar.selectbox("Select a BenchmarkFamily:",
                                     options=names)
df = df[df['name'] == selected_name]

# Get all unique columns from dataframe to allow user to select
cols = df.columns.tolist()

st.sidebar.write('Please select options for scatter plot')
group_by = st.sidebar.selectbox("Group results by:", options=cols, index=0)
x_axis = st.sidebar.selectbox("Select X-Axis:", options=cols, index=1)
y_axis = st.sidebar.selectbox("Select Y-Axis:", options=cols, index=2)
log_x = st.sidebar.checkbox("Log scale on X-axis", False)
log_y = st.sidebar.checkbox("Log scale on Y-axis", False)

pareto_curve = st.sidebar.checkbox("Display Pareto Curve", False)
x_better = st.sidebar.radio("Select the preferred direction for the X-axis",
                            options=["Higher is better", "Lower is better"],
                            index=1)
y_better = st.sidebar.radio("Select the preferred direction for the Y-axis",
                            options=["Higher is better", "Lower is better"],
                            index=1)

# Dynamic filtering based on column types
# Heading
st.sidebar.write('Please select filters for the data')
for col in df.columns:
    if np.issubdtype(df[col].dtype, np.number):
        min_val, max_val = df[col].min(), df[col].max()
        if min_val == max_val:
            # If min and max are the same, then the slider will not work
            # So we just show the value
            st.sidebar.write(f'{col} value: {min_val}')
        else:
            selected_range = st.sidebar.slider(
                f"{col} range", float(min_val), float(max_val),
                (float(min_val), float(max_val)))
            df = df[(df[col] >= selected_range[0])
                    & (df[col] <= selected_range[1])]
    elif pd.api.types.is_categorical_dtype(
            df[col].dtype) or pd.api.types.is_string_dtype(df[col].dtype):
        unique_vals = df[col].unique().tolist()
        selected_vals = st.sidebar.multiselect(f"{col} values",
                                               options=unique_vals,
                                               default=unique_vals)
        df = df[df[col].isin(selected_vals)]

# User can select runs to display
runs = df[group_by].unique().tolist()
selected_runs = st.sidebar.multiselect("Select runs to display:",
                                       options=runs,
                                       default=runs)

# Create grouped object
grouped = df[df[group_by].isin(selected_runs)].groupby(group_by)

fig, ax = plt.subplots()

# Scatter plot for each group
for name, group in grouped:
    ax.scatter(group[x_axis], group[y_axis], label=name, marker='x', alpha=0.5)

    if pareto_curve:
        pareto_points = compute_pareto_frontier(group, x_axis, y_axis,
                                                x_better, y_better)

        if pareto_points is not None:
            if x_better == "Lower is better":
                pareto_points[:, 0] = -pareto_points[:, 0]
            if y_better == "Lower is better":
                pareto_points[:, 1] = -pareto_points[:, 1]
            ax.plot(pareto_points[:, 0], pareto_points[:, 1])

if log_x:
    ax.set_xscale('log')
if log_y:
    ax.set_yscale('log')

ax.legend(loc='best')
plt.xlabel(x_axis)
plt.ylabel(y_axis)

st.pyplot(fig)



import json
import os
from dataclasses import dataclass
from typing import List

import numpy as np


@dataclass
class Result:
    system_info: dict
    solver: str
    options: dict
    parameters: dict
    results: dict


@dataclass
class BenchmarkFamily:
    name: str
    results: List[Result] = None

    @classmethod
    def from_dict(cls, name, data):
        results = []
        for result in data:
            results.append(Result(**result))
        return cls(name, results)


def load_results(directory):
    # Get the list of all json files in the directory
    json_files = [
        pos_json for pos_json in os.listdir(directory)
        if pos_json.endswith('.json')
    ]
    print(os.listdir(directory))

    # For each file
    results = []
    for json_file in json_files:
        # Load in a dictionary
        with open(os.path.join(directory, json_file)) as f:
            data = json.load(f)

        # Get the name of the benchmark family
        name = json_file.split('.')[0]

        # Create a BenchmarkFamily object
        benchmark_family = BenchmarkFamily.from_dict(name, data)

        # Add it to the list
        results.append(benchmark_family)

    return results


def compute_pareto_frontier(df,
                            x_axis,
                            y_axis,
                            x_better="Higher is better",
                            y_better="Higher is better"):
    """
    Compute the Pareto frontier for a given dataset and objectives
    """
    data_points = df[[x_axis, y_axis]].values

    if x_better == "Lower is better":
        data_points[:, 0] = -data_points[:, 0]

    if y_better == "Lower is better":
        data_points[:, 1] = -data_points[:, 1]

    # Order points by x_axis
    data_points = data_points[data_points[:, 0].argsort()]

    # Compute Pareto frontier
    pareto_frontier = [data_points[0]]
    for point in data_points[1:]:
        if point[1] >= pareto_frontier[-1][
                1]:    # Check if the point y value is greater than last point y
            pareto_frontier.append(point)

    return np.array(pareto_frontier)


def camel_to_snake_case(name):
    return ''.join(['_' + c.lower() if c.isupper() else c
                    for c in name]).lstrip('_')



import importlib.util
import itertools
import os

from .solver import Run


def load_modules_from_directory(directory):
    # Traverse directory recursively
    for root, dirs, files in os.walk(directory):
        for file in files:
            # Check if file is a Python file and starts in 'benchmark_'
            if file.endswith('.py') and file.startswith('benchmark_'):
                file_path = os.path.join(root, file)
                # Import module
                load_module_from_file(file_path)


def load_module_from_file(file_path):
    # Create module name from file path
    module_name = os.path.splitext(os.path.basename(file_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, file_path)

    module = importlib.util.module_from_spec(spec)

    # Run the module
    spec.loader.exec_module(module)


class RunRegister:
    runs = []

    def add_run(self, run):
        self.runs.append(run)

    def run_all(self, save_results=True):
        print(f"Running all {len(self.runs)} runs")

        while len(self.runs) > 0:
            run = self.runs.pop(0)
            print(f"Benchmark: {run.benchmark.name}")

            _ = run.run_benchmark()

            print(f"Results: {run._results}\n")

            if save_results:
                run.save_results()

            # Delete the run object
            del run

        print("All runs finished")


# Instantiate the register
run_register = RunRegister()


class BenchmarkRegistration:

    def __init__(self, func=None, parameters={}):
        self.func = func
        self.parameters = parameters

    def add_run(self):
        # Make run
        definition = self.func(**self.parameters)
        benchmark, solver = definition["benchmark"], definition["solver"]
        run = Run(benchmark, solver)

        print(f"Adding {benchmark.name} with {solver.name} to register")
        run_register.add_run(run)


def register_benchmark(vals):
    if isinstance(vals, tuple):
        func, parameters = vals
    else:
        func = vals
        parameters = None

    if parameters is None:
        registration = BenchmarkRegistration(func)
        registration.add_run()
    else:
        for params in parameters:
            registration = BenchmarkRegistration(func, params)
            registration.add_run()


def parametrize(param_name, values):

    def decorator(vals):
        if isinstance(vals, tuple):
            func, params = vals
        else:
            func = vals
            params = [{}]
        # the `params` is a list of dictionaries. Generate a new
        # list of dictionaries with the new parameter added as a key.
        # This should generate all combinations for the previous parameters
        # and the new one.
        new_params = []
        for value in values:
            new_entry = {param_name: value}
            params_update = params.copy()
            params_update = [dict(p, **new_entry) for p in params_update]
            new_params.extend(params_update)

        return func, new_params

    return decorator



import pathlib
import runpy
import sys

import streamlit.web.bootstrap


class Dashboard:

    def __init__(self, folder):
        self.folder = folder

    def run(self):
        # Get directory of this file
        this_dir = pathlib.Path(__file__).parent.absolute()
        script = this_dir / '_dashboard.py'

        sys.argv = ["streamlit", "run", script, f'--folder={self.folder}']
        runpy.run_module("streamlit", run_name="__main__")



# main.py
import argparse

from basil.dashboard import Dashboard
from basil.register import load_modules_from_directory, run_register


def main():
    parser = argparse.ArgumentParser(description='Process some inputs.')
    parser.add_argument('folder',
                        type=str,
                        help='folder containing the benchmark runs')
    parser.add_argument('--dashboard',
                        action='store_true',
                        help='run the dashboard')
    args = parser.parse_args()

    if args.dashboard:
        dashboard = Dashboard(args.folder)
        dashboard.run()
        return

    # Load modules from directory to register
    load_modules_from_directory(args.folder)

    # Run all benchmarks
    run_register.run_all()


if __name__ == '__main__':
    main()



import datetime
import json
import os
import platform
import subprocess

import psutil

from basil.utils import camel_to_snake_case


class Solver(object):

    def setup(self, inputs):
        pass

    @property
    def name(self):
        return camel_to_snake_case(self.__class__.__name__)

    @property
    def options(self):
        return {}

    def collect_system_info(self):
        info = {}

        # OS and Processor Info
        info["OS"] = platform.system()
        info["OS_version"] = platform.version()
        info["CPU"] = platform.processor()

        # RAM Info
        ram_info = psutil.virtual_memory()
        info["RAM"] = ram_info.total

        # GPU Info - this works best on Unix systems, may not be portable
        try:
            gpu_info = subprocess.check_output("lspci | grep ' VGA '",
                                               shell=True).decode("utf8")
            info["GPU"] = gpu_info
        except Exception as e:
            info["GPU"] = "No GPU info"

        return info


class Run:
    run_at = None
    sys_info = None
    benchmark = None
    solver = None
    _results = None

    def __init__(self, benchmark, solver):
        self.run_at = datetime.datetime.now()
        self.sys_info = solver.collect_system_info()
        self.benchmark = benchmark
        self.solver = solver

    def run_benchmark(self):
        # Generate inputs and setup solver
        input_and_aux = self.benchmark.inputs
        inputs, aux = input_and_aux["inputs"], input_and_aux["aux"]
        self.solver.setup(inputs)

        # Run benchmark
        outputs = self.solver.run(inputs)

        # Get results
        self.benchmark.validate_outputs(outputs)
        results = self.benchmark.evaluate(outputs, aux)

        # Add them to the object
        self.add_results(results)

        return results

    def add_results(self, results):
        # Construct the results dictionary
        self._results = {
            "system_info": self.sys_info,
            "solver": self.solver.name,
            "options": self.solver.options,
            "parameters": self.benchmark.parameters,
            "results": results
        }

    def save_results(self):
        import json
        import os

        benchmark_name = self.benchmark.name

        # Make sure the `results` folder exists, otherwise create it
        if not os.path.exists("results"):
            os.mkdir("results")

        # Check if the file exists, otherwise initialize it as
        # a list JSON object
        if not os.path.exists(f"results/{benchmark_name}.json"):
            with open(f"results/{benchmark_name}.json", "w") as f:
                f.write("[]")

        # Append the results to the file as a new entry of the list
        with open(f"results/{benchmark_name}.json", "r+") as f:
            # TODO: This method of loading and saving the file is not
            #      very efficient, should be probably changed by using
            #      a database
            results = json.load(f)
            results.append(self._results)
            # Write the results back to the file in a pretty format
            f.seek(0)
            json.dump(results, f, indent=2)
            f.truncate()

    def __call__(self):
        _ = self.run_benchmark()
        self.save_results()
        self.display_results()
